@@ -42,7 +42,9 @@ import adtlog
 
 known_restrictions = ['rw-build-tree', 'breaks-testbed', 'needs-root',
                       'build-needed', 'allow-stderr', 'isolation-container',
-                      'isolation-machine', 'needs-recommends', 'needs-reboot']
+                      'isolation-machine', 'needs-recommends', 'needs-reboot',
+                      'flaky', 'skippable', 'superficial',
+                      'skip-not-installable']
 
 
 class Unsupported(Exception):
@@ -80,7 +82,7 @@ class Test:
     actions.
     '''
     def __init__(self, name, path, command, restrictions, features, depends,
-                 clicks, installed_clicks):
+                 clicks, installed_clicks, synth_depends):
         '''Create new test description
 
         A test must have either "path" or "command", the respective other value
@@ -93,6 +95,7 @@ class Test:
         @depends: string list of test dependencies (packages)
         @clicks: path list of click packages to install for this test
         @installed_clicks: names of already installed clicks for this test
+        @synth_depends: string list of synthesized test dependencies (packages)
         '''
         if '/' in name:
             raise Unsupported(name, 'test name may not contain / character')
@@ -111,8 +114,10 @@ class Test:
         self.depends = depends
         self.clicks = clicks
         self.installed_clicks = installed_clicks
+        self.synth_depends = synth_depends
         # None while test hasn't run yet; True: pass, False: fail
         self.result = None
+        self.skipped = False
         adtlog.debug('Test defined: name %s path %s command "%s" '
                      'restrictions %s features %s depends %s clicks %s '
                      'installed clicks %s' %
@@ -123,13 +128,28 @@ class Test:
         '''Mark test as passed'''
 
         self.result = True
-        adtlog.report(self.name, 'PASS')
+        if 'superficial' in self.restrictions:
+            adtlog.report(self.name, 'PASS (superficial)')
+        else:
+            adtlog.report(self.name, 'PASS')
+
+    def set_skipped(self, reason):
+        '''Mark test as skipped'''
+        # This isn't called skipped() to avoid clashing with the boolean
+        # attribute.
+
+        self.skipped = True
+        self.result = True
+        adtlog.report(self.name, 'SKIP ' + reason)
 
     def failed(self, reason):
         '''Mark test as failed'''
 
         self.result = False
-        adtlog.report(self.name, 'FAIL ' + reason)
+        if 'flaky' in self.restrictions:
+            adtlog.report(self.name, 'FLAKY ' + reason)
+        else:
+            adtlog.report(self.name, 'FAIL ' + reason)
 
     def check_testbed_compat(self, caps):
         '''Check for restrictions incompatible with test bed capabilities.
@@ -221,6 +241,7 @@ def _debian_check_unknown_fields(name, record):
 
 def _debian_packages_from_source(srcdir):
     packages = []
+    packages_no_arch = []
 
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
         if 'Package' not in st:
@@ -232,11 +253,12 @@ def _debian_packages_from_source(srcdir):
             continue
         arch = st['Architecture']
         if arch in ('all', 'any'):
-            packages.append('%s (>= 0~)' % st['Package'])
+            packages.append(st['Package'])
         else:
-            packages.append('%s (>= 0~) [%s]' % (st['Package'], arch))
+            packages.append('%s [%s]' % (st['Package'], arch))
+        packages_no_arch.append(st['Package'])
 
-    return packages
+    return (packages, packages_no_arch)
 
 
 def _debian_build_deps_from_source(srcdir, testbed_arch):
@@ -246,6 +268,8 @@ def _debian_build_deps_from_source(srcdir, testbed_arch):
             deps += st['Build-depends']
         if 'Build-depends-indep' in st:
             deps += ', ' + st['Build-depends-indep']
+        if 'Build-depends-arch' in st:
+            deps += ', ' + st['Build-depends-arch']
 
     # resolve arch specific dependencies and build profiles
     perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
@@ -272,9 +296,9 @@ def _debian_build_deps_from_source(srcdir, testbed_arch):
 
 
 dep_re = re.compile(
-    r'(?P<package>[a-z0-9+-.]+)(?::native)?\s*'
+    r'(?P<package>[a-z0-9+-.]+)(?::[a-z0-9_-]+)?\s*'
     r'(\((?P<relation><<|<=|>=|=|>>)\s*(?P<version>[^\)]*)\))?'
-    r'(\s*\[[[a-z0-9+-.! ]+\])?$')
+    r'(\s*\[(?P<arch>[a-z0-9+-.! ]+)\])?$')
 
 
 def _debian_check_dep(testname, dep):
@@ -296,6 +320,38 @@ def _debian_check_dep(testname, dep):
             # too old python-debian, skip the check
             pass
 
+    return (m.group('package'), m.group('version'))
+
+
+def _synthesize_deps(dep, testbed_arch):
+    '''Ensure that apt can install synthesized Depends
+
+    Test Depends may have architecture qualifiers we need to check,
+    because apt command-line can't handle those and can't do the check.
+    (Policy says the architecture qualifier should not be in the binary
+    control file). We'll ignore the version here.
+    '''
+
+    dep = dep.strip()
+    m = dep_re.match(dep)
+
+    arch_matches = 'y'
+    if m.group('arch') is not None:
+        arch_matches = False
+        perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
+                                stdout=subprocess.PIPE)
+        code = '''use Dpkg::Deps;
+                  print deps_parse('foo [%s]')->arch_is_concerned('%s') ? 'y' : 'n';
+                  ''' % (m.group('arch'), testbed_arch)
+        arch_matches = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
+        if perl.returncode != 0:
+            raise InvalidControl('source', 'Invalid (Test-)Depends architecture qualifiers')
+
+    if arch_matches == 'y':
+        return m.group('package')
+    else:
+        return None
+
 
 def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
     '''Parse Depends: line in a Debian package
@@ -307,6 +363,8 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
     dependencies.
     '''
     deps = []
+    synthdeps = []
+    (my_packages, my_packages_no_arch) = _debian_packages_from_source(srcdir)
     for alt_group_str in dep_str.split(','):
         alt_group_str = alt_group_str.strip()
         if not alt_group_str:
@@ -314,19 +372,35 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
             continue
         adtlog.debug('processing dependency %s' % alt_group_str)
         if alt_group_str == '@':
-            for d in _debian_packages_from_source(srcdir):
+            for d in my_packages:
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
+                s = _synthesize_deps(d, testbed_arch)
+                if s:
+                    synthdeps.append(s)
         elif alt_group_str == '@builddeps@':
             for d in _debian_build_deps_from_source(srcdir, testbed_arch):
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
         else:
+            synthdep_alternatives = []
             for dep in alt_group_str.split('|'):
-                _debian_check_dep(testname, dep)
+                (pkg, version) = _debian_check_dep(testname, dep)
+                if pkg not in my_packages_no_arch:
+                    synthdep_alternatives = []
+                    break
+                s = _synthesize_deps(dep, testbed_arch)
+                if s:
+                    synthdep_alternatives.append(s)
+            if synthdep_alternatives:
+                    adtlog.debug('marked alternatives %s as a synthesised dependency' % synthdep_alternatives)
+                    if len(synthdep_alternatives) > 1:
+                        synthdeps.append(synthdep_alternatives)
+                    else:
+                        synthdeps.append(synthdep_alternatives[0])
             deps.append(alt_group_str)
 
-    return deps
+    return (deps, synthdeps)
 
 
 def _autodep8(srcdir):
@@ -373,16 +447,30 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
     tests = []
     if not control_path:
         control_path = os.path.join(srcdir, 'debian', 'tests', 'control')
+        dtc_exists = os.path.exists(control_path)
+        try_autodep8 = False
 
-        if not os.path.exists(control_path):
-            if auto_control:
-                control = _autodep8(srcdir)
-                if control is None:
-                    return ([], False)
-                control_path = control.name
+        if auto_control:
+            if not dtc_exists:
+                try_autodep8 = True
             else:
-                adtlog.debug('auto_control is disabled, no tests')
+                dcontrol_path = os.path.join(srcdir, 'debian', 'control')
+                for record in parse_rfc822(dcontrol_path):
+                    testsuite = record.get('Testsuite', '')
+                    if 'autopkgtest-pkg-' in testsuite:
+                        try_autodep8 = True
+                    # We only want to look at the source section
+                    break
+
+        if try_autodep8:
+            control = _autodep8(srcdir)
+            if control is not None:
+                control_path = control.name
+            elif not dtc_exists:
                 return ([], False)
+        elif not dtc_exists:
+            adtlog.debug('auto_control is disabled, and no regular tests')
+            return ([], False)
 
     for record in parse_rfc822(control_path):
         command = None
@@ -411,10 +499,11 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
 
             if 'Tests' in record:
                 test_names = record['Tests'].replace(',', ' ').split()
-                depends = _parse_debian_depends(test_names[0],
-                                                record.get('Depends', '@'),
-                                                srcdir,
-                                                testbed_arch)
+                (depends, synth_depends) = _parse_debian_depends(
+                    test_names[0],
+                    record.get('Depends', '@'),
+                    srcdir,
+                    testbed_arch)
                 if 'Test-command' in record:
                     raise InvalidControl('*', 'Only one of "Tests" or '
                                          '"Test-Command" may be given')
@@ -426,15 +515,16 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                 _debian_check_unknown_fields(test_names[0], record)
                 for n in test_names:
                     test = Test(n, os.path.join(test_dir, n), None,
-                                restrictions, features, depends, [], [])
+                                restrictions, features, depends, [], [], synth_depends)
                     test.check_testbed_compat(testbed_caps)
                     tests.append(test)
             elif 'Test-command' in record:
                 command = record['Test-command']
-                depends = _parse_debian_depends(command,
-                                                record.get('Depends', '@'),
-                                                srcdir,
-                                                testbed_arch)
+                (depends, synth_depends) = _parse_debian_depends(
+                    command,
+                    record.get('Depends', '@'),
+                    srcdir,
+                    testbed_arch)
                 if feature_test_name is None:
                     command_counter += 1
                     name = 'command%i' % command_counter
@@ -442,7 +532,7 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                     name = feature_test_name
                 _debian_check_unknown_fields(name, record)
                 test = Test(name, None, command, restrictions, features,
-                            depends, [], [])
+                            depends, [], [], synth_depends)
                 test.check_testbed_compat(testbed_caps)
                 tests.append(test)
             else:
@@ -532,7 +622,7 @@ def parse_click_manifest(manifest, testbed_caps, clickdeps, use_installed,
         try:
             test = Test(name, desc.get('path'), desc.get('command'),
                         desc.get('restrictions', []), desc.get('features', []),
-                        desc.get('depends', []), clickdeps, installed_clicks)
+                        desc.get('depends', []), clickdeps, installed_clicks, [])
             test.check_testbed_compat(testbed_caps)
             tests.append(test)
         except Unsupported as u:
@@ -545,7 +635,7 @@ def parse_click_manifest(manifest, testbed_caps, clickdeps, use_installed,
             try:
                 repo = manifest_j['x-source']['vcs-bzr']
                 adtlog.info('checking out click source from %s' % repo)
-                d = tempfile.mkdtemp(prefix='adt.clicksrc.')
+                d = tempfile.mkdtemp(prefix='autopkgtest.clicksrc.')
                 atexit.register(shutil.rmtree, d, ignore_errors=True)
                 try:
                     subprocess.check_call(['bzr', 'checkout', '--lightweight',
